# Autogenerated by devscripts/update-version.py

__version__ = '2026.07.04'

RELEASE_GIT_HEAD = '997fa140840a08df3938b40da470c78049fef1f6'

VARIANT = None

UPDATE_HINT = None

CHANNEL = 'stable'

ORIGIN = 'yt-dlp/yt-dlp'

_pkg_version = '2026.07.04'
